"""Python Tool Configuration."""
from __future__ import annotations

from dataclasses import dataclass  # Backwards compatibility

from .fields import field
from .pytoolconfig import PyToolConfig
from .types import UniversalKey

__all__ = ["PyToolConfig", "field", "UniversalKey", "dataclass"]



"""Source for pyproject.toml files or more generally toml files."""
from __future__ import annotations

import sys
from pathlib import Path

from pytoolconfig.sources.source import Source
from pytoolconfig.types import Key
from pytoolconfig.universal_config import UniversalConfig
from pytoolconfig.utils import (
    _dict_to_dataclass,
    find_config_file,
    max_py_version,
    min_py_version,
    parse_dependencies,
)

if sys.version_info < (3, 11, 0):
    import tomli as tomllib
else:
    import tomllib


class PyProject(Source):
    """Source for pyproject.toml and pytool.toml files.

    Can be extended to other toml files.
    """

    tool: str
    toml_dict: dict | None = None
    name: str = "pyproject.toml"
    description: str = """
    PEP 518 defines pyproject.toml as a configuration file to store build system
    requirements for Python projects. With the help of tools like Poetry or Flit
    it can fully replace the need for setup.py and setup.cfg files.
    """  # taken from black.
    file: Path | None

    def __init__(
        self,
        working_directory: Path,
        tool: str,
        bases: list[str] | None = None,
        recursive: bool = True,
    ):
        """Initialize the TOML configuration.

        :param working_directory: Working Directory
        :param tool: name of your tool. Will read configuration from [tool.yourtool]
        :param bases: Base files/folders to look for (besides pyproject.toml)
        :param recursive: search recursively up the directory tree for the file.
        """
        if recursive:
            self.file = find_config_file(working_directory, "pyproject.toml", bases)
        else:
            self.file = working_directory / "pyproject.toml"
        self.tool = tool

    def _read(self) -> bool:
        if not self.file or not self.file.exists():
            return False
        self.toml_dict = tomllib.loads(self.file.read_text())
        if self.toml_dict is None:
            return False
        if "tool" not in self.toml_dict.keys():
            return False
        return self.tool in self.toml_dict["tool"].keys()

    def parse(self) -> dict[str, Key] | None:
        """Parse the TOML file."""
        if not self._read():
            return None
        assert self.toml_dict
        return self.toml_dict["tool"][self.tool]

    def universalconfig(self) -> UniversalConfig:
        """Parse the file for the universal config object's fields.

        Only implement the relevant fields such as minimum python version.

        Pre: file was read but tool isn't necessarily in file.
        """
        if not self.toml_dict:
            return UniversalConfig()
        config: UniversalConfig
        if "pytoolconfig" in self.toml_dict.get("tool", {}).keys():
            config = _dict_to_dataclass(
                UniversalConfig, self.toml_dict["tool"]["pytoolconfig"]
            )
        else:
            config = UniversalConfig()
        if "project" in self.toml_dict.keys():
            project = self.toml_dict["project"]
            if "requires-python" in project.keys():
                raw_python_ver = project["requires-python"]
                config.min_py_version = min_py_version(raw_python_ver)
                config.max_py_version = max_py_version(raw_python_ver)
            if "dependencies" in project:
                dependencies = parse_dependencies(project["dependencies"])
                config.dependencies = list(dependencies)
            if "optional-dependencies" in project:
                optional_deps = {}
                for group, deps in project["optional-dependencies"].items():
                    optional_deps[group] = list(parse_dependencies(deps))
                config.optional_dependencies = optional_deps
            if "version" in project:
                config.version = project["version"]
        return config



"""Source for INI configuration files via configparser."""
from __future__ import annotations

from configparser import ConfigParser, SectionProxy
from pathlib import Path
from typing import Dict

from pytoolconfig.sources.source import Source
from pytoolconfig.types import Key
from pytoolconfig.utils import find_config_file


def _add_split_to_dict(
    dest: dict[str, Key], table_to_add: list[str], table: SectionProxy
) -> None:
    if len(table_to_add) == 0:
        for table_key in table:
            dest[table_key] = table[table_key]
    else:
        first = table_to_add[0]
        dest.setdefault(first, {})
        assert isinstance(dest[first], Dict)
        _add_split_to_dict(dest[first], table_to_add[1:], table)


class IniConfig(Source):
    """Source for INI configuration files via configparser."""

    _config: ConfigParser
    name: str
    description: str | None

    def __init__(
        self,
        working_directory: Path,
        filename: str,
        base_table: str,
        description: str | None = None,
    ):
        """Initialize the Ini Configuration.

        :param working_directory: the working directory to search.
        :param filename: the filename to search for.
        :param base_table: The table to search for.
            The file will only be used if this is present.
            The base_table will not be included in the parsed output.
        :param description: The description used in documentation.
        """
        self.file = find_config_file(working_directory, filename)
        self.base_table = base_table
        self.name = filename
        self.description = description
        self._config = ConfigParser()

    def _read(self) -> bool:
        if self.file is None:
            return False
        self._config.read_string(self.file.read_text())
        for table in self._config:
            split = table.split(".")
            if split[0] == self.base_table:
                return True
        return False

    def parse(self) -> dict[str, Key] | None:
        """Parse the INI file."""
        if not self._read():
            return None
        output: dict[str, Key] = {}
        for table in self._config:
            split = table.split(".")
            if split[0] == self.base_table:
                _add_split_to_dict(output, split[1:], self._config[table])
        return output



"""Program to generate documentation for a given PyToolConfig object."""


from __future__ import annotations

import sys
from dataclasses import is_dataclass
from typing import Any, Generator

from docutils.statemachine import StringList
from sphinx.application import Sphinx
from sphinx.ext.autodoc import ClassDocumenter
from tabulate import tabulate

from .fields import _gather_config_fields
from .pytoolconfig import PyToolConfig
from .sources import Source
from .types import ConfigField, Dataclass
from .universal_config import UniversalConfig

if sys.version_info < (3, 8, 0):
    from typing_extensions import get_origin
else:
    from typing import get_origin


def _type_to_str(type_to_print: type[Any]) -> str | None:
    if type_to_print is None:
        return None
    if get_origin(type_to_print) is None:
        try:
            return type_to_print.__name__
        except AttributeError:
            return type_to_print
    return str(type_to_print).replace("typing.", "")


def _subtables(model: type[Dataclass]) -> dict[str, type[Dataclass]]:
    result = {}
    for name, field in _gather_config_fields(model).items():
        if is_dataclass(field._type):
            result[name] = field._type
    return result


def _generate_table(
    model: type[Dataclass],
    tablefmt: str = "rst",
    prefix: str = "",
) -> Generator[str, None, None]:
    header = ["name", "description", "type", "default"]
    model_fields: dict[str, ConfigField] = _gather_config_fields(model)
    command_line = any(field.command_line for field in model_fields.values())
    if command_line:
        header.append("command line flag")
    table = []
    for name, field in model_fields.items():
        if not is_dataclass(field._type):
            row = [
                f"{name}" if prefix == "" else f"{prefix}.{name}",
                field.description.replace("\n", " ") if field.description else None,
                _type_to_str(field._type),
                field._default,
            ]
            if field.universal_config:
                key = field.universal_config
                assert is_dataclass(UniversalConfig)
                universal_key = _gather_config_fields(UniversalConfig)[key.name]
                row[1] = universal_key.description
                row[3] = universal_key._default
            if command_line:
                cli_doc = field.command_line
                if cli_doc is not None:
                    row.append(", ".join(cli_doc))
                else:
                    row.append(None)
            table.append(row)
    yield from tabulate(table, tablefmt=tablefmt, headers=header).split("\n")


class PyToolConfigAutoDocumenter(ClassDocumenter):
    """Sphinx autodocumenter for pytoolconfig models."""

    objtype = "pytoolconfigtable"
    content_indent = ""
    titles_allowed = True

    @classmethod
    def can_document_member(
        cls, member: Any, membername: str, isattr: bool, parent: Any
    ) -> bool:
        """Check if member is dataclass."""
        return is_dataclass(member)

    def add_directive_header(self, sig: str) -> None:
        """Remove directive headers."""

    def add_content(
        self, more_content: StringList | None, no_docstring: bool = False
    ) -> None:
        """Create simple table to document configuration options."""
        source = self.get_sourcename()
        config = self.object
        for line in _generate_table(config):
            self.add_line(line, source)


class PyToolConfigSourceDocumenter(ClassDocumenter):
    objtype = "pytoolconfigsources"
    content_indent = ""
    titles_allowed = True

    @classmethod
    def can_document_member(
        cls, member: Any, membername: str, isattr: bool, parent: Any
    ) -> bool:
        """Check if member is dataclass."""
        return isinstance(member, Source)

    def add_directive_header(self, sig: str) -> None:
        """Remove directive headers."""

    def add_content(
        self, more_content: StringList | None, no_docstring: bool = False
    ) -> None:
        """Create simple table to document configuration options."""
        source = self.get_sourcename()
        config = self.object
        for line in _generate_table(config):
            self.add_line(line, source)


def setup(app: Sphinx) -> None:
    """Register automatic documenter."""
    app.setup_extension("sphinx.ext.autodoc")
    app.add_autodocumenter(PyToolConfigAutoDocumenter)


def _generate_documentation(config: PyToolConfig) -> Generator[str, None, None]:
    """Generate Markdown documentation for a given config model.

    This currently Beta at best. Do not use.
    """
    yield "# Configuration\n"
    if len(config.sources) > 1:
        yield f"{config.tool} supports the following sources:\n"
        for idx, source in enumerate(config.sources):
            yield f" {idx}. {source.name}\n"
    else:
        name = next(config.sources).name
        yield f"{config.tool} supports the {name} format\n"
    yield "\n"
    for source in config.sources:
        if source.description:
            yield f"## {source.name} \n"
            yield source.description
            yield "\n"
    yield "## Options\n"
    yield from _generate_table(config.model, "github")
    yield "\n"
    for prefix, subtable in _subtables(config.model).items():
        yield from _generate_table(subtable, "github", prefix)
        yield "\n"
